import { EntryStorage, BrowserStorage } from "../models/storage";
import { Encryption } from "../models/encryption";
import * as CryptoJS from "crypto-js";
import { OTPType, OTPAlgorithm } from "../models/otp";
import { ActionContext } from "vuex";

export class Accounts implements Module {
  async getModule() {
    const cachedPassphrase = await this.getCachedPassphrase();
    const encryption: Encryption = new Encryption(cachedPassphrase);
    const shouldShowPassphrase = await EntryStorage.hasEncryptionKey();
    const entries = shouldShowPassphrase ? [] : await this.getEntries();

    return {
      state: {
        entries,
        encryption,
        OTPType,
        OTPAlgorithm,
        shouldShowPassphrase,
        sectorStart: false, // Should display timer circles?
        sectorOffset: 0, // Offset in seconds for animations
        second: 0, // Offset in seconds for math
        filter: true,
        siteName: await this.getSiteName(),
        showSearch: false,
        exportData: await EntryStorage.getExport(entries),
        exportEncData: await EntryStorage.getExport(entries, true),
        key: await BrowserStorage.getKey(),
        wrongPassword: false,
        initComplete: false,
      },
      getters: {
        shouldFilter(
          state: AccountsState,
          getters: { matchedEntries: string[] }
        ) {
          return (
            localStorage.smartFilter !== "false" &&
            getters.matchedEntries.length
          );
        },
        matchedEntries: (state: AccountsState) => {
          return this.matchedEntries(state.siteName, state.entries);
        },
        currentlyEncrypted(state: AccountsState) {
          for (const entry of state.entries) {
            if (entry.secret === null) {
              return true;
            }
          }
          return false;
        },
        entries(state: AccountsState) {
          const pinnedEntries = state.entries.filter((entry) => entry.pinned);
          const unpinnedEntries = state.entries.filter(
            (entry) => !entry.pinned
          );
          return [...pinnedEntries, ...unpinnedEntries];
        },
      },
      mutations: {
        stopFilter(state: AccountsState) {
          state.filter = false;
        },
        showSearch(state: AccountsState) {
          state.showSearch = true;
        },
        updateCodes(state: AccountsState) {
          let second = new Date().getSeconds();
          if (localStorage.offset) {
            // prevent second from negative
            second += Number(localStorage.offset) + 60;
          }

          second = second % 60;
          state.second = second;

          let currentlyEncrypted = false;

          for (const entry of state.entries) {
            if (entry.secret === null) {
              currentlyEncrypted = true;
            }
          }

          if (
            !state.sectorStart &&
            state.entries.length > 0 &&
            !currentlyEncrypted
          ) {
            state.sectorStart = true;
            state.sectorOffset = -second;
          }

          // if (second > 25) {
          //   app.class.timeout = true;
          // } else {
          //   app.class.timeout = false;
          // }
          // if (second < 1) {
          //   const entries = app.entries as OTP[];
          //   for (let i = 0; i < entries.length; i++) {
          //     if (entries[i].type !== OTPType.hotp &&
          //         entries[i].type !== OTPType.hhex) {
          //       entries[i].generate();
          //     }
          //   }
          // }
          const entries = state.entries as OTPEntryInterface[];
          for (let i = 0; i < entries.length; i++) {
            if (
              entries[i].type !== OTPType.hotp &&
              entries[i].type !== OTPType.hhex
            ) {
              entries[i].generate();
            }
          }
        },
        loadCodes(state: AccountsState, newCodes: OTPEntryInterface[]) {
          state.entries = newCodes;
        },
        moveCode(state: AccountsState, opts: { from: number; to: number }) {
          state.entries.splice(
            opts.to,
            0,
            state.entries.splice(opts.from, 1)[0]
          );

          for (let i = 0; i < state.entries.length; i++) {
            if (state.entries[i].index !== i) {
              state.entries[i].index = i;
            }
          }
        },
        pinEntry(state: AccountsState, entry: OTPEntryInterface) {
          state.entries[entry.index].pinned = !entry.pinned;
        },
        updateExport(
          state: AccountsState,
          exportData: { [k: string]: OTPEntryInterface }
        ) {
          state.exportData = exportData;
        },
        updateEncExport(
          state: AccountsState,
          exportData: { [k: string]: OTPEntryInterface }
        ) {
          state.exportEncData = exportData;
        },
        updateKeyExport(
          state: AccountsState,
          key: { enc: string; hash: string } | null
        ) {
          state.key = key;
        },
        wrongPassword(state: AccountsState) {
          state.wrongPassword = true;
        },
        initComplete(state: AccountsState) {
          state.initComplete = true;
        },
      },
      actions: {
        deleteCode: async (
          state: ActionContext<AccountsState, {}>,
          hash: string
        ) => {
          const index = state.state.entries.findIndex(
            (entry) => entry.hash === hash
          );
          if (index > -1) {
            state.state.entries.splice(index, 1);
          }
          state.commit(
            "updateExport",
            await EntryStorage.getExport(state.state.entries)
          );
          state.commit(
            "updateEncExport",
            await EntryStorage.getExport(state.state.entries, true)
          );
        },
        addCode: async (
          state: ActionContext<AccountsState, {}>,
          entry: OTPEntryInterface
        ) => {
          state.state.entries.unshift(entry);
          state.commit(
            "updateExport",
            await EntryStorage.getExport(state.state.entries)
          );
          state.commit(
            "updateEncExport",
            await EntryStorage.getExport(state.state.entries, true)
          );
        },
        applyPassphrase: async (
          state: ActionContext<AccountsState, {}>,
          password: string
        ) => {
          if (!password) {
            return;
          }

          state.commit("currentView/changeView", "LoadingPage", { root: true });

          const encKey = await BrowserStorage.getKey();
          if (!encKey) {
            // --- migrate to key
            // verify current password
            state.state.encryption.updateEncryptionPassword(password);
            await state.dispatch("updateEntries");

            if (state.getters.currentlyEncrypted) {
              state.commit("style/hideInfo", true, { root: true });
              return;
            }
            // gen key
            const wordArray = CryptoJS.lib.WordArray.random(120);
            const encKey = CryptoJS.AES.encrypt(wordArray, password).toString();
            const encKeyHash = await new Promise(
              (resolve: (value: string) => void) => {
                const iframe = document.getElementById("argon-sandbox");
                const message = {
                  action: "hash",
                  value: wordArray.toString(),
                };
                if (iframe) {
                  window.addEventListener("message", (response) => {
                    resolve(response.data.response);
                  });
                  // eslint-disable-next-line @typescript-eslint/ban-ts-ignore
                  //@ts-ignore
                  iframe.contentWindow.postMessage(message, "*");
                }
              }
            );

            if (!encKeyHash) {
              state.commit("style/hideInfo", true, { root: true });
              return;
            }

            // change entry encryption to key and remove old hash
            const oldKeys: string[] = [];
            for (const entry of state.state.entries) {
              await entry.changeEncryption(
                new Encryption(wordArray.toString())
              );
              oldKeys.push(entry.hash);
              entry.genUUID();
            }

            // store key
            await BrowserStorage.set({
              key: { enc: encKey, hash: encKeyHash },
            });
            await EntryStorage.set(state.state.entries);
            await new Promise((resolve) => {
              BrowserStorage.remove(oldKeys, () => {
                resolve();
              });
            });

            state.state.encryption.updateEncryptionPassword(
              wordArray.toString()
            );
            await state.dispatch("updateEntries");
          } else {
            // --- decrypt using key
            const key = CryptoJS.AES.decrypt(encKey.enc, password).toString();
            const isCorrectPassword = await new Promise(
              (resolve: (value: string) => void) => {
                const iframe = document.getElementById("argon-sandbox");
                const message = {
                  action: "verify",
                  value: key,
                  hash: encKey.hash,
                };
                if (iframe) {
                  window.addEventListener("message", (response) => {
                    resolve(response.data.response);
                  });
                  // eslint-disable-next-line @typescript-eslint/ban-ts-ignore
                  //@ts-ignore
                  iframe.contentWindow.postMessage(message, "*");
                }
              }
            );

            if (!isCorrectPassword) {
              state.commit("wrongPassword");
              state.commit("currentView/changeView", "EnterPasswordPage", {
                root: true,
              });
              return;
            }

            state.state.encryption.updateEncryptionPassword(key);
            await state.dispatch("updateEntries");

            // Encrypt any unencrypted entries.
            // Browser sync can cause unencrypted entries to show up.
            let needUpdateStorage = false;
            for (const entry of state.state.entries) {
              if (!entry.encSecret) {
                await entry.changeEncryption(state.state.encryption);
                needUpdateStorage = true;
              }
            }

            if (needUpdateStorage) {
              await EntryStorage.set(state.state.entries);
              await state.dispatch("updateEntries");
            }

            if (!state.getters.currentlyEncrypted) {
              chrome.runtime.sendMessage({
                action: "cachePassphrase",
                value: key,
              });
            }
          }
          state.commit("style/hideInfo", true, { root: true });
          return;
        },
        changePassphrase: async (
          state: ActionContext<AccountsState, {}>,
          password: string
        ) => {
          if (password) {
            const wordArray = CryptoJS.lib.WordArray.random(120);
            const encKey = CryptoJS.AES.encrypt(wordArray, password).toString();
            const encKeyHash = await new Promise(
              (resolve: (value: string) => void) => {
                const iframe = document.getElementById("argon-sandbox");
                const message = {
                  action: "hash",
                  value: wordArray.toString(),
                };
                if (iframe) {
                  window.addEventListener("message", (response) => {
                    resolve(response.data.response);
                  });
                  // eslint-disable-next-line @typescript-eslint/ban-ts-ignore
                  //@ts-ignore
                  iframe.contentWindow.postMessage(message, "*");
                }
              }
            );

            if (!encKeyHash) {
              return;
            }

            // change entry encryption and regen hash
            const removeHashes: string[] = [];
            for (const entry of state.state.entries) {
              await entry.changeEncryption(
                new Encryption(wordArray.toString())
              );
              // if not uuidv4 regen
              if (
                /[0-9A-F]{8}-[0-9A-F]{4}-4[0-9A-F]{3}-[89AB][0-9A-F]{3}-[0-9A-F]{12}/i.test(
                  entry.hash
                )
              ) {
                removeHashes.push(entry.hash);
                entry.genUUID();
              }
            }

            // store key
            await BrowserStorage.set({
              key: { enc: encKey, hash: encKeyHash },
            });
            await EntryStorage.set(state.state.entries);
            if (removeHashes.length) {
              await new Promise((resolve) => {
                BrowserStorage.remove(removeHashes, () => {
                  resolve();
                });
              });
            }

            state.state.encryption.updateEncryptionPassword(
              wordArray.toString()
            );

            await state.dispatch("updateEntries");

            // https://github.com/Authenticator-Extension/Authenticator/issues/412
            if (navigator.userAgent.indexOf("Chrome") !== -1) {
              await BrowserStorage.clearLogs();
            }

            chrome.runtime.sendMessage({
              action: "cachePassphrase",
              value: wordArray.toString(),
            });
          } else {
            for (const entry of state.state.entries) {
              await entry.changeEncryption(new Encryption(""));
            }
            await EntryStorage.set(state.state.entries);

            state.state.encryption.updateEncryptionPassword("");

            BrowserStorage.remove("key");

            await state.dispatch("updateEntries");

            chrome.runtime.sendMessage({
              action: "lock",
            });
          }

          // remove cached passphrase in old version
          localStorage.removeItem("encodedPhrase");
        },
        updateEntries: async (state: ActionContext<AccountsState, {}>) => {
          const entries = await this.getEntries();

          if (state.state.encryption.getEncryptionStatus()) {
            for (const entry of entries) {
              await entry.applyEncryption(state.state.encryption as Encryption);
            }
          }

          state.commit("loadCodes", entries);
          state.commit("updateCodes");
          state.commit(
            "updateExport",
            await EntryStorage.getExport(state.state.entries)
          );
          state.commit(
            "updateEncExport",
            await EntryStorage.getExport(state.state.entries, true)
          );
          state.commit("updateKeyExport", await BrowserStorage.getKey());
          state.commit("initComplete");
          return;
        },
        clearFilter: (state: ActionContext<AccountsState, {}>) => {
          state.commit("stopFilter");
          if (state.state.entries.length >= 10) {
            state.commit("showSearch");
          }
        },
        migrateStorage: async (
          state: ActionContext<AccountsState, {}>,
          newStorageLocation: string
        ) => {
          // sync => local
          if (
            localStorage.storageLocation === "sync" &&
            newStorageLocation === "local"
          ) {
            return new Promise((resolve, reject) => {
              chrome.storage.sync.get((syncData) => {
                chrome.storage.local.set(syncData, () => {
                  chrome.storage.local.get((localData) => {
                    // Double check if data was set
                    if (
                      Object.keys(syncData).every(
                        (value) => Object.keys(localData).indexOf(value) >= 0
                      )
                    ) {
                      localStorage.storageLocation = "local";
                      chrome.storage.sync.clear();
                      resolve("updateSuccess");
                      return;
                    } else {
                      reject(" All data not transferred successfully.");
                      return;
                    }
                  });
                });
              });
            });
            // local => sync
          } else if (
            localStorage.storageLocation === "local" &&
            newStorageLocation === "sync"
          ) {
            return new Promise((resolve, reject) => {
              chrome.storage.local.get((localData) => {
                chrome.storage.sync.set(localData, () => {
                  chrome.storage.sync.get((syncData) => {
                    // Double check if data was set
                    if (
                      Object.keys(localData).every(
                        (value) => Object.keys(syncData).indexOf(value) >= 0
                      )
                    ) {
                      localStorage.storageLocation = "sync";
                      chrome.storage.local.clear();
                      resolve("updateSuccess");
                      return;
                    } else {
                      reject(" All data not transferred successfully.");
                      return;
                    }
                  });
                });
              });
            });
          }
        },
      },
      namespaced: true,
    };
  }

  private async getSiteName() {
    return new Promise((resolve: (value: Array<string | null>) => void) => {
      chrome.tabs.query({ active: true, lastFocusedWindow: true }, (tabs) => {
        const tab = tabs[0];
        const query = new URLSearchParams(
          document.location.search.substring(1)
        );

        let title: string | null;
        let url: string | null;
        const titleFromQuery = query.get("title");
        const urlFromQuery = query.get("url");

        if (titleFromQuery && urlFromQuery) {
          title = decodeURIComponent(titleFromQuery);
          url = decodeURIComponent(urlFromQuery);
        } else {
          if (!tab) {
            return resolve([null, null]);
          }

          title = tab.title?.replace(/[^a-z0-9]/gi, "").toLowerCase() ?? null;
          url = tab.url ?? null;
        }

        if (!url) {
          return resolve([title, null]);
        }

        const urlParser = new URL(url);
        const hostname = urlParser.hostname; // it's always lower case

        // try to parse name from hostname
        // i.e. hostname is www.example.com
        // name should be example
        let nameFromDomain = "";

        // ip address
        if (/^\d+\.\d+\.\d+\.\d+$/.test(hostname)) {
          nameFromDomain = hostname;
        }

        // local network
        if (hostname.indexOf(".") === -1) {
          nameFromDomain = hostname;
        }

        const hostLevelUnits = hostname.split(".");

        if (hostLevelUnits.length === 2) {
          nameFromDomain = hostLevelUnits[0];
        }

        // www.example.com
        // example.com.cn
        if (hostLevelUnits.length > 2) {
          // example.com.cn
          if (
            ["com", "net", "org", "edu", "gov", "co"].indexOf(
              hostLevelUnits[hostLevelUnits.length - 2]
            ) !== -1
          ) {
            nameFromDomain = hostLevelUnits[hostLevelUnits.length - 3];
          } else {
            // www.example.com
            nameFromDomain = hostLevelUnits[hostLevelUnits.length - 2];
          }
        }

        nameFromDomain = nameFromDomain.replace(/-/g, "").toLowerCase();

        return resolve([title, nameFromDomain, hostname]);
      });
    });
  }

  private getCachedPassphrase() {
    return new Promise((resolve: (value: string) => void) => {
      chrome.runtime.sendMessage(
        { action: "passphrase" },
        (passphrase: string) => {
          return resolve(passphrase);
        }
      );
    });
  }

  private async getEntries() {
    const otpEntries = await EntryStorage.get();
    return otpEntries;
  }

  private matchedEntries(
    siteName: Array<string | null>,
    entries: OTPEntryInterface[]
  ) {
    if (siteName.length < 2) {
      return false;
    }

    const matched = [];

    for (const entry of entries) {
      if (this.isMatchedEntry(siteName, entry)) {
        matched.push(entry.hash);
      }
    }

    return matched;
  }

  private isMatchedEntry(
    siteName: Array<string | null>,
    entry: OTPEntryInterface
  ) {
    if (!entry.issuer) {
      return false;
    }

    const issuerHostMatches = entry.issuer.split("::");
    const issuer = issuerHostMatches[0]
      .replace(/[^0-9a-z]/gi, "")
      .toLowerCase();

    if (!issuer) {
      return false;
    }

    const siteTitle = siteName[0] || "";
    const siteNameFromHost = siteName[1] || "";
    const siteHost = siteName[2] || "";

    if (issuerHostMatches.length > 1) {
      if (siteHost && siteHost.indexOf(issuerHostMatches[1]) !== -1) {
        return true;
      }
    }
    // site title should be more detailed
    // so we use siteTitle.indexOf(issuer)
    if (siteTitle && siteTitle.indexOf(issuer) !== -1) {
      return true;
    }

    if (siteNameFromHost && issuer.indexOf(siteNameFromHost) !== -1) {
      return true;
    }

    return false;
  }
}
